#include <memory>
#include <string>
#include <string_view>
#include <unordered_map>

#include <grpcpp/grpcpp.h>
#include <openxr/openxr.h>
#include "render.grpc.pb.h"

#include "GUI/VRMenuObject.h"
#include "Render/BitmapFont.h"
#include "Render/GeometryBuilder.h"
#include "openxr_reflection_macro.h"
#include "XrApp.h"

#include "OVR_Math.h"
#include "Input/ControllerRenderer.h"
#include "Input/TinyUI.h"
#include "Render/GeometryRenderer.h"
#include "Render/SimpleBeamRenderer.h"

class RenderClient {
 public:
  RenderClient(std::shared_ptr<grpc::Channel> channel)
      : stub_(nerfviewer::Render::NewStub(channel)) {}

  // Assembles the client's payload, sends it and presents the response back
  // from the server.
  void Render(OVR::Matrix4f& pose, nerfviewer::Image* image) {

    // Construct transformation data
    nerfviewer::Transformation tf;
    for (int i = 0; i < 4; i++) {
        for (int j = 0; j < 4; j++) {
            tf.add_data(pose(i, j));
        }
    }

    // Context for the client. It could be used to convey extra information to
    // the server and/or tweak certain RPC behaviors.
    grpc::ClientContext context;

    // The actual RPC.
    grpc::Status status = stub_->Render(&context, tf, image);

    // Act upon its status.
    if (status.ok()) {
      ALOG("We got a reply");

    } else {
      ALOG("%d: %s", status.error_code(), status.error_message().c_str());
    }
  }

 private:
  std::unique_ptr<nerfviewer::Render::Stub> stub_;
};


class XrAppBaseApp : public OVRFW::XrApp {
   public:
    XrAppBaseApp() : OVRFW::XrApp() {
        BackgroundColor = OVR::Vector4f(0.55f, 0.35f, 0.1f, 1.0f);
    }

    virtual std::vector<const char*> GetExtensions() override {
        std::vector<const char*> extensions = XrApp::GetExtensions();

        // Add "XR_FB_color_space" to the list of extensions
        // The sample framework will add this to enabledExtensionNames
        // when calling xrCreateInstance()
        extensions.push_back(XR_FB_COLOR_SPACE_EXTENSION_NAME);
        return extensions;
    }

    virtual bool AppInit(const xrJava* context) override {
        // Init UI system
        if (false == ui_.Init(context, GetFileSys())) {
            ALOG("TinyUI::Init FAILED.");
            return false;
        }

        // Bind extension functions so we can call them
        OXR(xrGetInstanceProcAddr(
            GetInstance(),
            "xrEnumerateColorSpacesFB",
            reinterpret_cast<PFN_xrVoidFunction*>(&xrEnumerateColorSpacesFB)));
        OXR(xrGetInstanceProcAddr(
            GetInstance(),
            "xrSetColorSpaceFB",
            reinterpret_cast<PFN_xrVoidFunction*>(&xrSetColorSpaceFB)));

        // create grpc client
        std::string target_str = "localhost:50051";
        auto channel = grpc::CreateChannel(target_str, grpc::InsecureChannelCredentials());
        renderer_ = std::make_unique<RenderClient>(channel);
        ALOG("starting grpc client on %s", target_str.c_str());

        return true;
    }

    // Before this function, and after AppInit, XrApp::InitSession() calls:
    // - xrCreateSession(...) to create our Session
    // - xrCreateReferenceSpace(...) for local and stage space
    // - Create swapchain with xrCreateSwapchain(...)
    // - xrAttachSessionActionSets(...)
    // Before calling this function:
    virtual bool SessionInit() override {
        {
            // Get native HMD color space by chaining XrSystemColorSpacePropertiesFB
            // to the next pointer of XrSystemProperties
            XrSystemColorSpacePropertiesFB colorSpaceProps{
                XR_TYPE_SYSTEM_COLOR_SPACE_PROPERTIES_FB};
            XrSystemProperties sysprops{XR_TYPE_SYSTEM_PROPERTIES};
            sysprops.next = &colorSpaceProps;
            OXR(xrGetSystemProperties(GetInstance(), GetSystemId(), &sysprops));

            // The sample just displays the native HMD color space,
            // but an app could use this to load different assets
            // and to inform which colorspace it wants to select.
            std::string str =
                "Native HMD color space: " + std::string(XrEnumStr(colorSpaceProps.colorSpace));
            ui_.AddLabel(str, {-1.0f, 2.4f, -2.0f}, {700.0f, 45.0f});
        }

        // The default source color space is not something that's
        // provided by XrColorSpaceFB OpenXR extension, so just initialize
        // the label to say [default]
        declaredLabel_ = ui_.AddLabel(
            "App declared color space: [default]", {-1.0f, 2.3f, -2.0f}, {700.0f, 45.0f});

        { // Enumerate supported source color spaces with xrEnumerateColorSpaceFB

            // xrEnumerateColorSpacesFB uses the two-call idiom
            // First call with nullptr to get the required size:
            uint32_t colorSpaceCount{0};
            OXR(xrEnumerateColorSpacesFB(GetSession(), 0, &colorSpaceCount, nullptr));

            // Allocate array with enough space for the data
            auto colorSpaces = std::vector<XrColorSpaceFB>(colorSpaceCount);

            // Second call gets the actual data
            OXR(xrEnumerateColorSpacesFB(
                GetSession(), (uint32_t)colorSpaces.size(), &colorSpaceCount, colorSpaces.data()));

            // Header
            ui_.AddLabel("Compatible color spaces:", {-1.0f, 2.15f, -2.0f}, {700.0f, 45.0f});

            // Create one button for each supported color space conversion
            float posY = 2.0f;
            for (auto colorSpace : colorSpaces) {
                ui_.AddButton(XrEnumStr(colorSpace), {-1.0f, posY, -2.0f}, {600.0f, 60.0f}, [=]() {
                    // Set source color space with xrSetColorSpaceFB when button is pressed
                    // Remember that this is the source colorspace and no the target colorspace
                    OXR(xrSetColorSpaceFB(GetSession(), colorSpace));

                    declaredColorSpace_ = colorSpace; // For display in the UI
                });
                posY -= 0.15f;
            }
        }


        /// Disable scene navitgation
        GetScene().SetFootPos({0.0f, 0.0f, 0.0f});
        this->FreeMove = false;

        // Init objects that need OpenXR Session
        if (false == controllerRenderL_.Init(true)) {
            ALOG("SessionInit::Init L controller renderer FAILED.");
            return false;
        }
        if (false == controllerRenderR_.Init(false)) {
            ALOG("SessionInit::Init R controller renderer FAILED.");
            return false;
        }
        cursorBeamRenderer_.Init(GetFileSys(), nullptr, OVR::Vector4f(1.0f), 1.0f);

        return true;
    }

    // Called every frame
    virtual void Update(const OVRFW::ovrApplFrameIn& in) override {
        // Update the UI with the latest declared color space
        if (declaredColorSpace_ != XR_COLOR_SPACE_MAX_ENUM_FB) {
            std::string label = "App declared color space: ";
            label += XrEnumStr(declaredColorSpace_);
            declaredLabel_->SetText(label.c_str());
        }

//        colorCheckerRenderer_.Update();

        OVR::Matrix4f pose(in.HeadPose);
        nerfviewer::Image image;
        renderer_->Render(pose, &image);
        // debug info
//        char buf[1000];
//        tf.ToString(buf, 1000);
//        ALOG("tf is: %s", buf);


        // Clear the intersection rays from last frame:
        ui_.HitTestDevices().clear();

        if (in.LeftRemoteTracked) {
            controllerRenderL_.Update(in.LeftRemotePose);
            const bool didTrigger = in.LeftRemoteIndexTrigger > 0.5f;
            ui_.AddHitTestRay(in.LeftRemotePointPose, didTrigger);
        }

        if (in.RightRemoteTracked) {
            controllerRenderR_.Update(in.RightRemotePose);
            const bool didTrigger = in.RightRemoteIndexTrigger > 0.5f;
            ui_.AddHitTestRay(in.RightRemotePointPose, didTrigger);
        }

        ui_.Update(in);
        cursorBeamRenderer_.Update(in, ui_.HitTestDevices());
    }

    virtual void Render(const OVRFW::ovrApplFrameIn& in, OVRFW::ovrRendererOutput& out) override {
        colorCheckerRenderer_.Render(out.Surfaces);

        ui_.Render(in, out);

        if (in.LeftRemoteTracked) {
            controllerRenderL_.Render(out.Surfaces);
        }
        if (in.RightRemoteTracked) {
            controllerRenderR_.Render(out.Surfaces);
        }

        /// Render beams last, since they render with transparency
        cursorBeamRenderer_.Render(in, out);
    }

    virtual void SessionEnd() override {
        controllerRenderL_.Shutdown();
        controllerRenderR_.Shutdown();
        cursorBeamRenderer_.Shutdown();
        colorCheckerRenderer_.Shutdown();
    }

    virtual void AppShutdown(const xrJava* context) override {
        OVRFW::XrApp::AppShutdown(context);
        ui_.Shutdown();
    }

   private:
    // Function pointers to extension functions
    // Remember that you need to call xrGetInstanceProcAddr after
    // instance initiation to get these
    PFN_xrEnumerateColorSpacesFB xrEnumerateColorSpacesFB{nullptr};
    PFN_xrSetColorSpaceFB xrSetColorSpaceFB{nullptr};

    XrColorSpaceFB declaredColorSpace_{XR_COLOR_SPACE_MAX_ENUM_FB};

    OVRFW::VRMenuObject* declaredLabel_;
    OVRFW::GeometryRenderer colorCheckerRenderer_;
    OVRFW::ControllerRenderer controllerRenderL_;
    OVRFW::ControllerRenderer controllerRenderR_;
    OVRFW::TinyUI ui_;

    // Renderer that draws the beam from the controller
    OVRFW::SimpleBeamRenderer cursorBeamRenderer_;
    std::vector<OVRFW::ovrBeamRenderer::handle_t> beams_;

    // grpc client
    std::unique_ptr<RenderClient> renderer_;
};

ENTRY_POINT(XrAppBaseApp)
